/**
 * Copyright (c) 2012 - 2019 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.gecko.core.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.Executors;

import org.gecko.core.tests.ServiceCheckerCustomizer;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceException;
import org.osgi.framework.ServiceReference;
import org.osgi.util.promise.Promise;
import org.osgi.util.promise.PromiseFactory;

/**
 * 
 * @author mark
 * @since 25.11.2019
 */
@RunWith(MockitoJUnitRunner.class)
public class ServiceCustomizerTest {
	
	@Mock
	private BundleContext ctx;
	@Mock
	private ServiceReference<String> ref1;
	private ServiceCheckerCustomizer<String, String> customizer;
	private PromiseFactory pf = new PromiseFactory(Executors.newFixedThreadPool(3), Executors.newScheduledThreadPool(3));
	
	@Before
	public void setup() {
		customizer = new ServiceCheckerCustomizer<String, String>(ctx, pf);
	}
	
	@SuppressWarnings("unchecked")
	@Test
	public void testCustomizer() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		Mockito.when(ctx.getService(Mockito.any(ServiceReference.class))).thenReturn("Test");
		customizer.addingService(ref1);
		customizer.modifiedService(ref1, new String("Test"));
		customizer.removedService(ref1, new String("Test"));
		
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveCreate(1, 0);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(1, 1000);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(2, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(1, 0);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(1, 1000);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(2, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(1, 0);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(1, 1000);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(2, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
	}
	
	@SuppressWarnings("unchecked")
	@Test
	public void testCustomizerCreate_One() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		Mockito.when(ctx.getService(Mockito.any(ServiceReference.class))).thenReturn("Test");
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveCreate(1, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(1, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		// Service added with timeout
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(1, 1000);
		customizer.addingService(ref1);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 1000);
		
		// Service added with expectation smaller that current add value
		promise = customizer.resolveCreate(0, 1000);
		customizer.addingService(ref1);
		assertEquals(Boolean.FALSE, promise.getValue());
	}
	
	/**
	 * This use-case occurs, if a component is getting activated but not finished it.
	 * So there is no service but a service reference
	 * @throws InvocationTargetException
	 * @throws InterruptedException
	 */
	@SuppressWarnings("unchecked")
	@Test
	public void testCustomizerCreate_ReferenceNoService() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		Mockito.when(ctx.getService(Mockito.any(ServiceReference.class))).thenReturn(null);
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveCreate(1, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(1, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		// Service added with timeout
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(0, 1000);
		customizer.addingService(ref1);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
	}
	
	/**
	 * This use-case occurs, if a component is getting activated but not finished it.
	 * If there is an exception in the activate method, there can be an exception in the get service method
	 * @throws InvocationTargetException
	 * @throws InterruptedException
	 */
	@SuppressWarnings("unchecked")
	@Test
	public void testCustomizerCreate_GetServiceException() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		Mockito.when(ctx.getService(Mockito.any(ServiceReference.class))).thenThrow(ServiceException.class);
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveCreate(1, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(1, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		// Service added with timeout
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(0, 1000);
		customizer.addingService(ref1);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
	}
	
	@SuppressWarnings("unchecked")
	@Test
	public void testCustomizerCreate_Many() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		long duration = System.currentTimeMillis();
		Mockito.when(ctx.getService(Mockito.any(ServiceReference.class))).thenReturn("Test");
		Promise<Boolean> promise = customizer.resolveCreate(3, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(3, 1000);
		customizer.addingService(ref1);
		customizer.addingService(ref1);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		// Service added with timeout
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(3, 1000);
		customizer.addingService(ref1);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 1000);
	}
	
	@Test
	public void testCustomizerModify_Many() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveModify(3, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(3, 1000);
		customizer.modifiedService(ref1, new String("Test"));
		customizer.modifiedService(ref1, new String("Test"));
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		// Service added with timeout
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(3, 1000);
		customizer.modifiedService(ref1, new String("Test"));
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 1000);
	}
	
	@Test
	public void testCustomizerRemove_Many() throws InvocationTargetException, InterruptedException {
		// Nothing added with and without timeout
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveRemove(3, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(3, 1000);
		customizer.removedService(ref1, new String("Test"));
		customizer.removedService(ref1, new String("Test"));
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		// Service added with timeout
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(3, 1000);
		customizer.removedService(ref1, new String("Test"));
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 1000);
	}
	
	@Test
	public void testCustomizerCreate_Nothing() throws InvocationTargetException, InterruptedException {
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveCreate(0, 0);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 100);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveCreate(0, 1000);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
	}
	
	@Test
	public void testCustomizerModify_One() throws InvocationTargetException, InterruptedException {
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveModify(1, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(1, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(1, 1000);
		customizer.modifiedService(ref1, new String("Test"));
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 1000);
	}
	
	@Test
	public void testCustomizerModify_Nothing() throws InvocationTargetException, InterruptedException {
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveModify(0, 0);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 100);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveModify(0, 1000);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
	}
	
	@Test
	public void testCustomizerRemove_One() throws InvocationTargetException, InterruptedException {
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveRemove(1, 0);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 500);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(1, 1000);
		assertEquals(Boolean.FALSE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(1, 1000);
		customizer.removedService(ref1, new String("Test"));
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 1000);
	}
	
	@Test
	public void testCustomizerRemove_Nothing() throws InvocationTargetException, InterruptedException {
		long duration = System.currentTimeMillis();
		Promise<Boolean> promise = customizer.resolveRemove(0, 0);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration < 100);
		
		duration = System.currentTimeMillis();
		promise = customizer.resolveRemove(0, 1000);
		assertEquals(Boolean.TRUE, promise.getValue());
		duration = System.currentTimeMillis() - duration;
		assertTrue(duration >= 1000);
	}

}
